"""
The `pydantic_extra_types.ULID` module provides the [`ULID`] data type.

This class depends on the [python-ulid] package, which is a validate by the [ULID-spec](https://github.com/ulid/spec#implementations-in-other-languages).
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Union

from pydantic import GetCoreSchemaHandler
from pydantic._internal import _repr
from pydantic_core import PydanticCustomError, core_schema

try:
    from ulid import ULID as _ULID
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError(
        'The `ulid` module requires "python-ulid" to be installed. You can install it with "pip install python-ulid".'
    )

UlidType = Union[str, bytes, int]


@dataclass
class ULID(_repr.Representation):
    """
    A wrapper around [python-ulid](https://pypi.org/project/python-ulid/) package, which
    is a validate by the [ULID-spec](https://github.com/ulid/spec#implementations-in-other-languages).
    """

    ulid: _ULID

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.no_info_wrap_validator_function(
            cls._validate_ulid,
            core_schema.union_schema(
                [
                    core_schema.is_instance_schema(_ULID),
                    core_schema.int_schema(),
                    core_schema.bytes_schema(),
                    core_schema.str_schema(),
                ]
            ),
        )

    @classmethod
    def _validate_ulid(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Any:
        ulid: _ULID
        try:
            if isinstance(value, int):
                ulid = _ULID.from_int(value)
            elif isinstance(value, str):
                ulid = _ULID.from_str(value)
            elif isinstance(value, _ULID):
                ulid = value
            else:
                ulid = _ULID.from_bytes(value)
        except ValueError:
            raise PydanticCustomError('ulid_format', 'Unrecognized format')
        return handler(ulid)



"""
The `pydantic_extra_types.coordinate` module provides the [`Latitude`][pydantic_extra_types.coordinate.Latitude],
[`Longitude`][pydantic_extra_types.coordinate.Longitude], and
[`Coordinate`][pydantic_extra_types.coordinate.Coordinate] data types.
"""
from dataclasses import dataclass
from typing import Any, ClassVar, Tuple, Type

from pydantic import GetCoreSchemaHandler
from pydantic._internal import _repr
from pydantic_core import ArgsKwargs, PydanticCustomError, core_schema


class Latitude(float):
    """Latitude value should be between -90 and 90, inclusive.

    ```py
    from pydantic import BaseModel
    from pydantic_extra_types.coordinate import Latitude

    class Location(BaseModel):
        latitude: Latitude

    location = Location(latitude=41.40338)
    print(location)
    #> latitude=41.40338
    ```
    """

    min: ClassVar[float] = -90.00
    max: ClassVar[float] = 90.00

    @classmethod
    def __get_pydantic_core_schema__(cls, source: Type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.float_schema(ge=cls.min, le=cls.max)


class Longitude(float):
    """Longitude value should be between -180 and 180, inclusive.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.coordinate import Longitude

    class Location(BaseModel):
        longitude: Longitude

    location = Location(longitude=2.17403)
    print(location)
    #> longitude=2.17403
    ```
    """

    min: ClassVar[float] = -180.00
    max: ClassVar[float] = 180.00

    @classmethod
    def __get_pydantic_core_schema__(cls, source: Type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.float_schema(ge=cls.min, le=cls.max)


@dataclass
class Coordinate(_repr.Representation):
    """Coordinate parses Latitude and Longitude.

    You can use the `Coordinate` data type for storing coordinates. Coordinates can be
    defined using one of the following formats:

    1. Tuple: `(Latitude, Longitude)`. For example: `(41.40338, 2.17403)`.
    2. `Coordinate` instance: `Coordinate(latitude=Latitude, longitude=Longitude)`.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.coordinate import Coordinate

    class Location(BaseModel):
        coordinate: Coordinate

    location = Location(coordinate=(41.40338, 2.17403))
    #> coordinate=Coordinate(latitude=41.40338, longitude=2.17403)
    ```
    """

    _NULL_ISLAND: ClassVar[Tuple[float, float]] = (0.0, 0.0)

    latitude: Latitude
    longitude: Longitude

    @classmethod
    def __get_pydantic_core_schema__(cls, source: Type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        schema_chain = [
            core_schema.no_info_wrap_validator_function(cls._parse_str, core_schema.str_schema()),
            core_schema.no_info_wrap_validator_function(
                cls._parse_tuple,
                handler.generate_schema(Tuple[float, float]),
            ),
            handler(source),
        ]

        chain_length = len(schema_chain)
        chain_schemas = [core_schema.chain_schema(schema_chain[x:]) for x in range(chain_length - 1, -1, -1)]
        return core_schema.no_info_wrap_validator_function(
            cls._parse_args, core_schema.union_schema(chain_schemas)  # type: ignore[arg-type]
        )

    @classmethod
    def _parse_args(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Any:
        if isinstance(value, ArgsKwargs) and not value.kwargs:
            n_args = len(value.args)
            if n_args == 0:
                value = cls._NULL_ISLAND
            elif n_args == 1:
                value = value.args[0]
        return handler(value)

    @classmethod
    def _parse_str(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Any:
        if not isinstance(value, str):
            return value
        try:
            value = tuple(float(x) for x in value.split(','))
        except ValueError:
            raise PydanticCustomError(
                'coordinate_error',
                'value is not a valid coordinate: string is not recognized as a valid coordinate',
            )
        return ArgsKwargs(args=value)

    @classmethod
    def _parse_tuple(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Any:
        if not isinstance(value, tuple):
            return value
        return ArgsKwargs(args=handler(value))

    def __str__(self) -> str:
        return f'{self.latitude},{self.longitude}'

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Coordinate) and self.latitude == other.latitude and self.longitude == other.longitude

    def __hash__(self) -> int:
        return hash((self.latitude, self.longitude))



"""
The `pydantic_extra_types.phone_numbers` module provides the
[`PhoneNumber`][pydantic_extra_types.phone_numbers.PhoneNumber] data type.

This class depends on the [phonenumbers] package, which is a Python port of Google's [libphonenumber].
"""
from __future__ import annotations

from typing import Any, Callable, ClassVar, Generator

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic_core import PydanticCustomError, core_schema

try:
    import phonenumbers
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError(
        '`PhoneNumber` requires "phonenumbers" to be installed. You can install it with "pip install phonenumbers"'
    )

GeneratorCallableStr = Generator[Callable[..., str], None, None]


class PhoneNumber(str):
    """
    A wrapper around [phonenumbers](https://pypi.org/project/phonenumbers/) package, which
    is a Python port of Google's [libphonenumber](https://github.com/google/libphonenumber/).
    """

    supported_regions: list[str] = sorted(phonenumbers.SUPPORTED_REGIONS)
    """The supported regions."""
    supported_formats: list[str] = sorted([f for f in phonenumbers.PhoneNumberFormat.__dict__.keys() if f.isupper()])
    """The supported phone number formats."""

    default_region_code: ClassVar[str | None] = None
    """The default region code to use when parsing phone numbers without an international prefix."""
    phone_format: str = 'RFC3966'
    """The format of the phone number."""
    min_length: int = 7
    """The minimum length of the phone number."""
    max_length: int = 64
    """The maximum length of the phone number."""

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        json_schema = handler(schema)
        json_schema.update({'format': 'phone'})
        return json_schema

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(min_length=cls.min_length, max_length=cls.max_length),
        )

    @classmethod
    def _validate(cls, phone_number: str, _: core_schema.ValidationInfo) -> str:
        try:
            parsed_number = phonenumbers.parse(phone_number, cls.default_region_code)
        except phonenumbers.phonenumberutil.NumberParseException as exc:
            raise PydanticCustomError('value_error', 'value is not a valid phone number') from exc
        if not phonenumbers.is_valid_number(parsed_number):
            raise PydanticCustomError('value_error', 'value is not a valid phone number')

        return phonenumbers.format_number(parsed_number, getattr(phonenumbers.PhoneNumberFormat, cls.phone_format))



from datetime import datetime, timezone
from typing import Any

import pytest
from pydantic import BaseModel, ValidationError

from pydantic_extra_types.ulid import ULID

try:
    from ulid import ULID as _ULID
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError(
        'The `ulid` module requires "python-ulid" to be installed. You can install it with "pip install python-ulid".'
    )


class Something(BaseModel):
    ulid: ULID


@pytest.mark.parametrize(
    'ulid, result, valid',
    [
        # Valid ULID for str format
        ('01BTGNYV6HRNK8K8VKZASZCFPE', '01BTGNYV6HRNK8K8VKZASZCFPE', True),
        ('01BTGNYV6HRNK8K8VKZASZCFPF', '01BTGNYV6HRNK8K8VKZASZCFPF', True),
        # Invalid ULID for str format
        ('01BTGNYV6HRNK8K8VKZASZCFP', None, False),  # Invalid ULID (short length)
        ('01BTGNYV6HRNK8K8VKZASZCFPEA', None, False),  # Invalid ULID (long length)
        # Valid ULID for _ULID format
        (_ULID.from_str('01BTGNYV6HRNK8K8VKZASZCFPE'), '01BTGNYV6HRNK8K8VKZASZCFPE', True),
        (_ULID.from_str('01BTGNYV6HRNK8K8VKZASZCFPF'), '01BTGNYV6HRNK8K8VKZASZCFPF', True),
        # Invalid _ULID for bytes format
        (b'\x01\xBA\x1E\xB2\x8A\x9F\xFAy\x10\xD5\xA5k\xC8', None, False),  # Invalid ULID (short length)
        (b'\x01\xBA\x1E\xB2\x8A\x9F\xFAy\x10\xD5\xA5k\xC8\xB6\x00', None, False),  # Invalid ULID (long length)
        # Valid ULID for int format
        (109667145845879622871206540411193812282, '2JG4FVY7N8XS4GFVHPXGJZ8S9T', True),
        (109667145845879622871206540411193812283, '2JG4FVY7N8XS4GFVHPXGJZ8S9V', True),
        (109667145845879622871206540411193812284, '2JG4FVY7N8XS4GFVHPXGJZ8S9W', True),
    ],
)
def test_format_for_ulid(ulid: Any, result: Any, valid: bool):
    if valid:
        assert str(Something(ulid=ulid).ulid) == result
    else:
        with pytest.raises(ValidationError, match='format'):
            Something(ulid=ulid)


def test_property_for_ulid():
    ulid = Something(ulid='01BTGNYV6HRNK8K8VKZASZCFPE').ulid
    assert ulid.hex == '015ea15f6cd1c56689a373fab3f63ece'
    assert ulid == '01BTGNYV6HRNK8K8VKZASZCFPE'
    assert ulid.datetime == datetime(2017, 9, 20, 22, 18, 59, 153000, tzinfo=timezone.utc)
    assert ulid.timestamp == 1505945939.153


def test_json_schema():
    assert Something.model_json_schema(mode='validation') == {
        'properties': {
            'ulid': {
                'anyOf': [{'type': 'integer'}, {'format': 'binary', 'type': 'string'}, {'type': 'string'}],
                'title': 'Ulid',
            }
        },
        'required': ['ulid'],
        'title': 'Something',
        'type': 'object',
    }
    assert Something.model_json_schema(mode='serialization') == {
        'properties': {
            'ulid': {
                'anyOf': [{'type': 'integer'}, {'format': 'binary', 'type': 'string'}, {'type': 'string'}],
                'title': 'Ulid',
            }
        },
        'required': ['ulid'],
        'title': 'Something',
        'type': 'object',
    }



from typing import Any

import pytest
from pydantic import BaseModel, ValidationError

from pydantic_extra_types.phone_numbers import PhoneNumber


class Something(BaseModel):
    phone_number: PhoneNumber


# Note: the 555 area code will result in an invalid phone number
def test_valid_phone_number() -> None:
    Something(phone_number='+1 901 555 1212')


def test_when_extension_provided() -> None:
    Something(phone_number='+1 901 555 1212 ext 12533')


@pytest.mark.parametrize('invalid_number', ['', '123', 12, None, object(), '55 121'])
def test_invalid_phone_number(invalid_number: Any) -> None:
    with pytest.raises(ValidationError):
        Something(phone_number=invalid_number)


def test_formats_phone_number() -> None:
    result = Something(phone_number='+1 901 555 1212 ext 12533')
    assert result.phone_number == 'tel:+1-901-555-1212;ext=12533'


def test_supported_regions() -> None:
    assert 'US' in PhoneNumber.supported_regions
    assert 'GB' in PhoneNumber.supported_regions


def test_supported_formats() -> None:
    assert 'E164' in PhoneNumber.supported_formats
    assert 'RFC3966' in PhoneNumber.supported_formats
    assert '__dict__' not in PhoneNumber.supported_formats
    assert 'to_string' not in PhoneNumber.supported_formats


def test_parse_error() -> None:
    with pytest.raises(ValidationError, match='value is not a valid phone number'):
        Something(phone_number='555 1212')


def test_parsed_but_not_a_valid_number() -> None:
    with pytest.raises(ValidationError, match='value is not a valid phone number'):
        Something(phone_number='+1 555-1212')


def test_json_schema() -> None:
    assert Something.model_json_schema() == {
        'title': 'Something',
        'type': 'object',
        'properties': {
            'phone_number': {
                'title': 'Phone Number',
                'type': 'string',
                'format': 'phone',
                'minLength': 7,
                'maxLength': 64,
            }
        },
        'required': ['phone_number'],
    }



from typing import Any

import pytest
from pydantic import BaseModel, ValidationError

from pydantic_extra_types.routing_number import ABARoutingNumber


class Model(BaseModel):
    routing_number: ABARoutingNumber


@pytest.mark.parametrize('routing_number', [12, None, object(), 123456789])
def test_invalid_routing_number_string(routing_number: Any) -> None:
    with pytest.raises(ValidationError) as validation_error:
        Model(routing_number=routing_number)
    assert validation_error.match('Input should be a valid string')


@pytest.mark.parametrize('routing_number', ['', '123', '1234567890'])
def test_invalid_routing_number_length(routing_number: Any) -> None:
    with pytest.raises(ValidationError) as validation_error:
        Model(routing_number=routing_number)
    assert validation_error.match(r'String should have at (most|least) 9 characters')


@pytest.mark.parametrize('routing_number', ['122105154', '122235822', '123103723', '074900781'])
def test_invalid_routing_number(routing_number: Any) -> None:
    with pytest.raises(ValidationError) as validation_error:
        Model(routing_number=routing_number)
    assert validation_error.match('Incorrect ABA routing transit number')


@pytest.mark.parametrize('routing_number', ['122105155', '122235821', '123103729', '074900783'])
def test_valid_routing_number(routing_number: str) -> None:
    Model(routing_number=routing_number)


def test_raises_error_when_not_a_string() -> None:
    with pytest.raises(ValidationError, match='routing number is not all digits'):
        Model(routing_number='A12210515')



"""
The `pydantic_extra_types.mac_address` module provides functionality to parse and validate MAC addresses in different
formats, such as IEEE 802 MAC-48, EUI-48, EUI-64, or a 20-octet format.
"""

from __future__ import annotations

from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


class MacAddress(str):
    """Represents a MAC address and provides methods for conversion, validation, and serialization.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.mac_address import MacAddress


    class Network(BaseModel):
        mac_address: MacAddress


    network = Network(mac_address="00:00:5e:00:53:01")
    print(network)
    #> mac_address='00:00:5e:00:53:01'
    ```
    """

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.with_info_before_validator_function(
            cls._validate,
            core_schema.str_schema(),
        )

    @classmethod
    def _validate(cls, __input_value: str, _: Any) -> str:
        return cls.validate_mac_address(__input_value.encode())

    @staticmethod
    def validate_mac_address(value: bytes) -> str:
        """
        Validate a MAC Address from the provided byte value.

        Args:
            value: The byte value representing the MAC address.

        Returns:
            str: The parsed MAC address.

        Raises:
            PydanticCustomError: If the value is not a valid MAC address.
        """
        if len(value) < 14:
            raise PydanticCustomError(
                'mac_address_len',
                'Length for a {mac_address} MAC address must be {required_length}',
                {'mac_address': value.decode(), 'required_length': 14},
            )

        if value[2] in [ord(':'), ord('-')]:
            if (len(value) + 1) % 3 != 0:
                raise PydanticCustomError(
                    'mac_address_format', 'Must have the format xx:xx:xx:xx:xx:xx or xx-xx-xx-xx-xx-xx'
                )
            n = (len(value) + 1) // 3
            if n not in (6, 8, 20):
                raise PydanticCustomError(
                    'mac_address_format',
                    'Length for a {mac_address} MAC address must be {required_length}',
                    {'mac_address': value.decode(), 'required_length': (6, 8, 20)},
                )
            mac_address = bytearray(n)
            x = 0
            for i in range(n):
                try:
                    byte_value = int(value[x : x + 2], 16)
                    mac_address[i] = byte_value
                    x += 3
                except ValueError as e:
                    raise PydanticCustomError('mac_address_format', 'Unrecognized format') from e

        elif value[4] == ord('.'):
            if (len(value) + 1) % 5 != 0:
                raise PydanticCustomError('mac_address_format', 'Must have the format xx.xx.xx.xx.xx.xx')
            n = 2 * (len(value) + 1) // 5
            if n not in (6, 8, 20):
                raise PydanticCustomError(
                    'mac_address_format',
                    'Length for a {mac_address} MAC address must be {required_length}',
                    {'mac_address': value.decode(), 'required_length': (6, 8, 20)},
                )
            mac_address = bytearray(n)
            x = 0
            for i in range(0, n, 2):
                try:
                    byte_value = int(value[x : x + 2], 16)
                    mac_address[i] = byte_value
                    byte_value = int(value[x + 2 : x + 4], 16)
                    mac_address[i + 1] = byte_value
                    x += 5
                except ValueError as e:
                    raise PydanticCustomError('mac_address_format', 'Unrecognized format') from e

        else:
            raise PydanticCustomError('mac_address_format', 'Unrecognized format')

        return ':'.join(f'{b:02x}' for b in mac_address)



import pytest
from pydantic import BaseModel

from pydantic_extra_types.color import Color
from pydantic_extra_types.coordinate import Coordinate, Latitude, Longitude
from pydantic_extra_types.country import (
    CountryAlpha2,
    CountryAlpha3,
    CountryNumericCode,
    CountryShortName,
)
from pydantic_extra_types.isbn import ISBN
from pydantic_extra_types.mac_address import MacAddress
from pydantic_extra_types.payment import PaymentCardNumber
from pydantic_extra_types.ulid import ULID


@pytest.mark.parametrize(
    'cls,expected',
    [
        (
            Color,
            {
                'properties': {'x': {'format': 'color', 'title': 'X', 'type': 'string'}},
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            PaymentCardNumber,
            {
                'properties': {
                    'x': {
                        'maxLength': 19,
                        'minLength': 12,
                        'title': 'X',
                        'type': 'string',
                    }
                },
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            CountryAlpha2,
            {
                'properties': {'x': {'pattern': '^\\w{2}$', 'title': 'X', 'type': 'string'}},
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            CountryAlpha3,
            {
                'properties': {'x': {'pattern': '^\\w{3}$', 'title': 'X', 'type': 'string'}},
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            CountryNumericCode,
            {
                'properties': {'x': {'pattern': '^[0-9]{3}$', 'title': 'X', 'type': 'string'}},
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            CountryShortName,
            {
                'properties': {'x': {'title': 'X', 'type': 'string'}},
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            MacAddress,
            {
                'properties': {
                    'x': {
                        'title': 'X',
                        'type': 'string',
                    }
                },
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            Latitude,
            {
                'properties': {
                    'x': {
                        'maximum': 90.0,
                        'minimum': -90.0,
                        'title': 'X',
                        'type': 'number',
                    }
                },
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            Longitude,
            {
                'properties': {
                    'x': {
                        'maximum': 180.0,
                        'minimum': -180.0,
                        'title': 'X',
                        'type': 'number',
                    }
                },
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            Coordinate,
            {
                '$defs': {
                    'Coordinate': {
                        'properties': {
                            'latitude': {'maximum': 90.0, 'minimum': -90.0, 'title': 'Latitude', 'type': 'number'},
                            'longitude': {'maximum': 180.0, 'minimum': -180.0, 'title': 'Longitude', 'type': 'number'},
                        },
                        'required': ['latitude', 'longitude'],
                        'title': 'Coordinate',
                        'type': 'object',
                    }
                },
                'properties': {
                    'x': {
                        'anyOf': [
                            {'$ref': '#/$defs/Coordinate'},
                            {
                                'maxItems': 2,
                                'minItems': 2,
                                'prefixItems': [
                                    {'type': 'number'},
                                    {'type': 'number'},
                                ],
                                'type': 'array',
                            },
                            {'type': 'string'},
                        ],
                        'title': 'X',
                    },
                },
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            ULID,
            {
                'properties': {
                    'x': {
                        'anyOf': [{'type': 'integer'}, {'format': 'binary', 'type': 'string'}, {'type': 'string'}],
                        'title': 'X',
                    }
                },
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
        (
            ISBN,
            {
                'properties': {
                    'x': {
                        'title': 'X',
                        'type': 'string',
                    }
                },
                'required': ['x'],
                'title': 'Model',
                'type': 'object',
            },
        ),
    ],
)
def test_json_schema(cls, expected):
    class Model(BaseModel):
        x: cls

    assert Model.model_json_schema() == expected



from typing import Any

import pytest
from pydantic import BaseModel, ValidationError

from pydantic_extra_types.isbn import ISBN


class Book(BaseModel):
    isbn: ISBN


isbn_length_test_cases = [
    # Valid ISBNs
    ('8537809667', '9788537809662', True),  # ISBN-10 as input
    ('9788537809662', '9788537809662', True),  # ISBN-13 as input
    ('080442957X', '9780804429573', True),  # ISBN-10 ending in "X" as input
    ('9788584390670', '9788584390670', True),  # ISBN-13 Starting with 978
    ('9790306406156', '9790306406156', True),  # ISBN-13 starting with 979
    # Invalid ISBNs
    ('97885843906701', None, False),  # Length: 14 (Higher)
    ('978858439067', None, False),  # Length: 12 (In Between)
    ('97885843906', None, False),  # Length: 11 (In Between)
    ('978858439', None, False),  # Length: 9 (Lower)
    ('', None, False),  # Length: 0 (Lower)
]


@pytest.mark.parametrize('input_isbn, output_isbn, valid', isbn_length_test_cases)
def test_isbn_length(input_isbn: Any, output_isbn: str, valid: bool) -> None:
    if valid:
        assert Book(isbn=ISBN(input_isbn)).isbn == output_isbn
    else:
        with pytest.raises(ValidationError, match='isbn_length'):
            Book(isbn=ISBN(input_isbn))


isbn10_digits_test_cases = [
    # Valid ISBNs
    ('8537809667', '9788537809662', True),  # ISBN-10 as input
    ('080442957X', '9780804429573', True),  # ISBN-10 ending in "X" as input
    # Invalid ISBNs
    ('@80442957X', None, False),  # Non Integer in [0] position
    ('8@37809667', None, False),  # Non Integer in [1] position
    ('85@7809667', None, False),  # Non Integer in [2] position
    ('853@809667', None, False),  # Non Integer in [3] position
    ('8537@09667', None, False),  # Non Integer in [4] position
    ('85378@9667', None, False),  # Non Integer in [5] position
    ('853780@667', None, False),  # Non Integer in [6] position
    ('8537809@67', None, False),  # Non Integer in [7] position
    ('85378096@7', None, False),  # Non Integer in [8] position
    ('853780966@', None, False),  # Non Integer or X in [9] position
]


@pytest.mark.parametrize('input_isbn, output_isbn, valid', isbn10_digits_test_cases)
def test_isbn10_digits(input_isbn: Any, output_isbn: str, valid: bool) -> None:
    if valid:
        assert Book(isbn=ISBN(input_isbn)).isbn == output_isbn
    else:
        with pytest.raises(ValidationError, match='isbn10_invalid_characters'):
            Book(isbn=ISBN(input_isbn))


isbn13_digits_test_cases = [
    # Valid ISBNs
    ('9788537809662', '9788537809662', True),  # ISBN-13 as input
    ('9780306406157', '9780306406157', True),  # ISBN-13 as input
    ('9788584390670', '9788584390670', True),  # ISBN-13 Starting with 978
    ('9790306406156', '9790306406156', True),  # ISBN-13 starting with 979
    # Invalid ISBNs
    ('@788537809662', None, False),  # Non Integer in [0] position
    ('9@88537809662', None, False),  # Non Integer in [1] position
    ('97@8537809662', None, False),  # Non Integer in [2] position
    ('978@537809662', None, False),  # Non Integer in [3] position
    ('9788@37809662', None, False),  # Non Integer in [4] position
    ('97885@7809662', None, False),  # Non Integer in [5] position
    ('978853@809662', None, False),  # Non Integer in [6] position
    ('9788537@09662', None, False),  # Non Integer in [7] position
    ('97885378@9662', None, False),  # Non Integer in [8] position
    ('978853780@662', None, False),  # Non Integer in [9] position
    ('9788537809@62', None, False),  # Non Integer in [10] position
    ('97885378096@2', None, False),  # Non Integer in [11] position
    ('978853780966@', None, False),  # Non Integer in [12] position
]


@pytest.mark.parametrize('input_isbn, output_isbn, valid', isbn13_digits_test_cases)
def test_isbn13_digits(input_isbn: Any, output_isbn: str, valid: bool) -> None:
    if valid:
        assert Book(isbn=ISBN(input_isbn)).isbn == output_isbn
    else:
        with pytest.raises(ValidationError, match='isbn13_invalid_characters'):
            Book(isbn=ISBN(input_isbn))


isbn13_early_digits_test_cases = [
    # Valid ISBNs
    ('9780306406157', '9780306406157', True),  # ISBN-13 as input
    ('9788584390670', '9788584390670', True),  # ISBN-13 Starting with 978
    ('9790306406156', '9790306406156', True),  # ISBN-13 starting with 979
    # Invalid ISBNs
    ('1788584390670', None, False),  # Does not start with 978 or 979
    ('9288584390670', None, False),  # Does not start with 978 or 979
    ('9738584390670', None, False),  # Does not start with 978 or 979
]


@pytest.mark.parametrize('input_isbn, output_isbn, valid', isbn13_early_digits_test_cases)
def test_isbn13_early_digits(input_isbn: Any, output_isbn: str, valid: bool) -> None:
    if valid:
        assert Book(isbn=ISBN(input_isbn)).isbn == output_isbn
    else:
        with pytest.raises(ValidationError, match='isbn_invalid_early_characters'):
            Book(isbn=ISBN(input_isbn))


isbn_last_digit_test_cases = [
    # Valid ISBNs
    ('8537809667', '9788537809662', True),  # ISBN-10 as input
    ('9788537809662', '9788537809662', True),  # ISBN-13 as input
    ('080442957X', '9780804429573', True),  # ISBN-10 ending in "X" as input
    ('9788584390670', '9788584390670', True),  # ISBN-13 Starting with 978
    ('9790306406156', '9790306406156', True),  # ISBN-13 starting with 979
    # Invalid ISBNs
    ('8537809663', None, False),  # ISBN-10 as input with wrong last digit
    ('9788537809661', None, False),  # ISBN-13 as input with wrong last digit
    ('080442953X', None, False),  # ISBN-10 ending in "X" as input with wrong last digit
    ('9788584390671', None, False),  # ISBN-13 Starting with 978 with wrong last digit
    ('9790306406155', None, False),  # ISBN-13 starting with 979 with wrong last digit
]


@pytest.mark.parametrize('input_isbn, output_isbn, valid', isbn_last_digit_test_cases)
def test_isbn_last_digit(input_isbn: Any, output_isbn: str, valid: bool) -> None:
    if valid:
        assert Book(isbn=ISBN(input_isbn)).isbn == output_isbn
    else:
        with pytest.raises(ValidationError, match='isbn_invalid_digit_check_isbn'):
            Book(isbn=ISBN(input_isbn))


isbn_conversion_test_cases = [
    # Valid ISBNs
    ('8537809667', '9788537809662'),
    ('080442957X', '9780804429573'),
    ('9788584390670', '9788584390670'),
    ('9790306406156', '9790306406156'),
]


@pytest.mark.parametrize('input_isbn, output_isbn', isbn_conversion_test_cases)
def test_isbn_conversion(input_isbn: Any, output_isbn: str) -> None:
    assert Book(isbn=ISBN(input_isbn)).isbn == output_isbn



from re import Pattern
from typing import Any, Optional

import pytest
from pydantic import BaseModel, ValidationError
from pydantic_core._pydantic_core import ArgsKwargs

from pydantic_extra_types.coordinate import Coordinate, Latitude, Longitude


class Coord(BaseModel):
    coord: Coordinate


class Lat(BaseModel):
    lat: Latitude


class Lng(BaseModel):
    lng: Longitude


@pytest.mark.parametrize(
    'coord, result, error',
    [
        # Valid coordinates
        ((20.0, 10.0), (20.0, 10.0), None),
        ((-90.0, 0.0), (-90.0, 0.0), None),
        (('20.0', 10.0), (20.0, 10.0), None),
        ((20.0, '10.0'), (20.0, 10.0), None),
        ((45.678, -123.456), (45.678, -123.456), None),
        (('45.678, -123.456'), (45.678, -123.456), None),
        (Coordinate(20.0, 10.0), (20.0, 10.0), None),
        (Coordinate(latitude=0, longitude=0), (0, 0), None),
        (ArgsKwargs(args=()), (0, 0), None),
        (ArgsKwargs(args=(1, 0.0)), (1.0, 0), None),
        # # Invalid coordinates
        ((), None, 'Field required'),  # Empty tuple
        ((10.0,), None, 'Field required'),  # Tuple with only one value
        (('ten, '), None, 'string is not recognized as a valid coordinate'),
        ((20.0, 10.0, 30.0), None, 'Tuple should have at most 2 items'),  # Tuple with more than 2 values
        (ArgsKwargs(args=(1.0,)), None, 'Input should be a dictionary or an instance of Coordinate'),
        (
            '20.0, 10.0, 30.0',
            None,
            'Input should be a dictionary or an instance of Coordinate ',
        ),  # Str with more than 2 values
        ('20.0, 10.0, 30.0', None, 'Unexpected positional argument'),  # Str with more than 2 values
        (2, None, 'Input should be a dictionary or an instance of Coordinate'),  # Wrong type
    ],
)
def test_format_for_coordinate(coord: (Any, Any), result: (float, float), error: Optional[Pattern]):
    if error is None:
        _coord: Coordinate = Coord(coord=coord).coord
        print('vars(_coord)', vars(_coord))
        assert _coord.latitude == result[0]
        assert _coord.longitude == result[1]
    else:
        with pytest.raises(ValidationError, match=error):
            Coord(coord=coord).coord


@pytest.mark.parametrize(
    'coord, error',
    [
        # Valid coordinates
        ((-90.0, 0.0), None),
        ((50.0, 180.0), None),
        # Invalid coordinates
        ((-91.0, 0.0), 'Input should be greater than or equal to -90'),
        ((50.0, 181.0), 'Input should be less than or equal to 180'),
    ],
)
def test_limit_for_coordinate(coord: (Any, Any), error: Optional[Pattern]):
    if error is None:
        _coord: Coordinate = Coord(coord=coord).coord
        assert _coord.latitude == coord[0]
        assert _coord.longitude == coord[1]
    else:
        with pytest.raises(ValidationError, match=error):
            Coord(coord=coord).coord


@pytest.mark.parametrize(
    'latitude, valid',
    [
        # Valid latitude
        (20.0, True),
        (3.0000000000000000000000, True),
        (90.0, True),
        ('90.0', True),
        (-90.0, True),
        ('-90.0', True),
        # Unvalid latitude
        (91.0, False),
        (-91.0, False),
    ],
)
def test_format_latitude(latitude: float, valid: bool):
    if valid:
        _lat = Lat(lat=latitude).lat
        assert _lat == float(latitude)
    else:
        with pytest.raises(ValidationError, match='1 validation error for Lat'):
            Lat(lat=latitude)


@pytest.mark.parametrize(
    'longitude, valid',
    [
        # Valid latitude
        (20.0, True),
        (3.0000000000000000000000, True),
        (90.0, True),
        ('90.0', True),
        (-90.0, True),
        ('-90.0', True),
        (91.0, True),
        (-91.0, True),
        (180.0, True),
        (-180.0, True),
        # Unvalid latitude
        (181.0, False),
        (-181.0, False),
    ],
)
def test_format_longitude(longitude: float, valid: bool):
    if valid:
        _lng = Lng(lng=longitude).lng
        assert _lng == float(longitude)
    else:
        with pytest.raises(ValidationError, match='1 validation error for Lng'):
            Lng(lng=longitude)


def test_str_repr():
    assert str(Coord(coord=(20.0, 10.0)).coord) == '20.0,10.0'
    assert str(Coord(coord=('20.0, 10.0')).coord) == '20.0,10.0'
    assert repr(Coord(coord=(20.0, 10.0)).coord) == 'Coordinate(latitude=20.0, longitude=10.0)'


def test_eq():
    assert Coord(coord=(20.0, 10.0)).coord != Coord(coord='20.0,11.0').coord
    assert Coord(coord=('20.0, 10.0')).coord != Coord(coord='20.0,11.0').coord
    assert Coord(coord=('20.0, 10.0')).coord != Coord(coord='20.0,11.0').coord
    assert Coord(coord=(20.0, 10.0)).coord == Coord(coord='20.0,10.0').coord


def test_hashable():
    assert hash(Coord(coord=(20.0, 10.0)).coord) == hash(Coord(coord=(20.0, 10.0)).coord)
    assert hash(Coord(coord=(20.0, 11.0)).coord) != hash(Coord(coord=(20.0, 10.0)).coord)


def test_json_schema():
    class Model(BaseModel):
        value: Coordinate

    assert Model.model_json_schema(mode='validation')['$defs']['Coordinate'] == {
        'properties': {
            'latitude': {'maximum': 90.0, 'minimum': -90.0, 'title': 'Latitude', 'type': 'number'},
            'longitude': {'maximum': 180.0, 'minimum': -180.0, 'title': 'Longitude', 'type': 'number'},
        },
        'required': ['latitude', 'longitude'],
        'title': 'Coordinate',
        'type': 'object',
    }
    assert Model.model_json_schema(mode='validation')['properties']['value'] == {
        'anyOf': [
            {'$ref': '#/$defs/Coordinate'},
            {
                'maxItems': 2,
                'minItems': 2,
                'prefixItems': [{'type': 'number'}, {'type': 'number'}],
                'type': 'array',
            },
            {'type': 'string'},
        ],
        'title': 'Value',
    }
    assert Model.model_json_schema(mode='serialization') == {
        '$defs': {
            'Coordinate': {
                'properties': {
                    'latitude': {'maximum': 90.0, 'minimum': -90.0, 'title': 'Latitude', 'type': 'number'},
                    'longitude': {'maximum': 180.0, 'minimum': -180.0, 'title': 'Longitude', 'type': 'number'},
                },
                'required': ['latitude', 'longitude'],
                'title': 'Coordinate',
                'type': 'object',
            }
        },
        'properties': {'value': {'allOf': [{'$ref': '#/$defs/Coordinate'}], 'title': 'Value'}},
        'required': ['value'],
        'title': 'Model',
        'type': 'object',
    }



"""
Country definitions that are based on the [ISO 3166](https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes).
"""
from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from typing import Any

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic_core import PydanticCustomError, core_schema

try:
    import pycountry
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError(
        'The `country` module requires "pycountry" to be installed. You can install it with "pip install pycountry".'
    )


@dataclass
class CountryInfo:
    alpha2: str
    alpha3: str
    numeric_code: str
    short_name: str


@lru_cache
def _countries() -> list[CountryInfo]:
    return [
        CountryInfo(
            alpha2=country.alpha_2,
            alpha3=country.alpha_3,
            numeric_code=country.numeric,
            short_name=country.name,
        )
        for country in pycountry.countries
    ]


@lru_cache
def _index_by_alpha2() -> dict[str, CountryInfo]:
    return {country.alpha2: country for country in _countries()}


@lru_cache
def _index_by_alpha3() -> dict[str, CountryInfo]:
    return {country.alpha3: country for country in _countries()}


@lru_cache
def _index_by_numeric_code() -> dict[str, CountryInfo]:
    return {country.numeric_code: country for country in _countries()}


@lru_cache
def _index_by_short_name() -> dict[str, CountryInfo]:
    return {country.short_name: country for country in _countries()}


class CountryAlpha2(str):
    """CountryAlpha2 parses country codes in the [ISO 3166-1 alpha-2](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2)
    format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.country import CountryAlpha2

    class Product(BaseModel):
        made_in: CountryAlpha2

    product = Product(made_in='ES')
    print(product)
    #> made_in='ES'
    ```
    """

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> CountryAlpha2:
        if __input_value not in _index_by_alpha2():
            raise PydanticCustomError('country_alpha2', 'Invalid country alpha2 code')
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(to_upper=True),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        json_schema = handler(schema)
        json_schema.update({'pattern': r'^\w{2}$'})
        return json_schema

    @property
    def alpha3(self) -> str:
        """The country code in the [ISO 3166-1 alpha-3](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3) format."""
        return _index_by_alpha2()[self].alpha3

    @property
    def numeric_code(self) -> str:
        """The country code in the [ISO 3166-1 numeric](https://en.wikipedia.org/wiki/ISO_3166-1_numeric) format."""
        return _index_by_alpha2()[self].numeric_code

    @property
    def short_name(self) -> str:
        """The country short name."""
        return _index_by_alpha2()[self].short_name


class CountryAlpha3(str):
    """CountryAlpha3 parses country codes in the [ISO 3166-1 alpha-3](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3)
    format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.country import CountryAlpha3

    class Product(BaseModel):
        made_in: CountryAlpha3

    product = Product(made_in="USA")
    print(product)
    #> made_in='USA'
    ```
    """

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> CountryAlpha3:
        if __input_value not in _index_by_alpha3():
            raise PydanticCustomError('country_alpha3', 'Invalid country alpha3 code')
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(to_upper=True),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        json_schema = handler(schema)
        json_schema.update({'pattern': r'^\w{3}$'})
        return json_schema

    @property
    def alpha2(self) -> str:
        """The country code in the [ISO 3166-1 alpha-2](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) format."""
        return _index_by_alpha3()[self].alpha2

    @property
    def numeric_code(self) -> str:
        """The country code in the [ISO 3166-1 numeric](https://en.wikipedia.org/wiki/ISO_3166-1_numeric) format."""
        return _index_by_alpha3()[self].numeric_code

    @property
    def short_name(self) -> str:
        """The country short name."""
        return _index_by_alpha3()[self].short_name


class CountryNumericCode(str):
    """CountryNumericCode parses country codes in the
    [ISO 3166-1 numeric](https://en.wikipedia.org/wiki/ISO_3166-1_numeric) format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.country import CountryNumericCode

    class Product(BaseModel):
        made_in: CountryNumericCode

    product = Product(made_in="840")
    print(product)
    #> made_in='840'
    ```
    """

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> CountryNumericCode:
        if __input_value not in _index_by_numeric_code():
            raise PydanticCustomError('country_numeric_code', 'Invalid country numeric code')
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(to_upper=True),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        json_schema = handler(schema)
        json_schema.update({'pattern': r'^[0-9]{3}$'})
        return json_schema

    @property
    def alpha2(self) -> str:
        """The country code in the [ISO 3166-1 alpha-2](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) format."""
        return _index_by_numeric_code()[self].alpha2

    @property
    def alpha3(self) -> str:
        """The country code in the [ISO 3166-1 alpha-3](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3) format."""
        return _index_by_numeric_code()[self].alpha3

    @property
    def short_name(self) -> str:
        """The country short name."""
        return _index_by_numeric_code()[self].short_name


class CountryShortName(str):
    """CountryShortName parses country codes in the short name format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.country import CountryShortName

    class Product(BaseModel):
        made_in: CountryShortName

    product = Product(made_in="United States")
    print(product)
    #> made_in='United States'
    ```
    """

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> CountryShortName:
        if __input_value not in _index_by_short_name():
            raise PydanticCustomError('country_short_name', 'Invalid country short name')
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @property
    def alpha2(self) -> str:
        """The country code in the [ISO 3166-1 alpha-2](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) format."""
        return _index_by_short_name()[self].alpha2

    @property
    def alpha3(self) -> str:
        """The country code in the [ISO 3166-1 alpha-3](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3) format."""
        return _index_by_short_name()[self].alpha3

    @property
    def numeric_code(self) -> str:
        """The country code in the [ISO 3166-1 numeric](https://en.wikipedia.org/wiki/ISO_3166-1_numeric) format."""
        return _index_by_short_name()[self].numeric_code



"""
The `pydantic_extra_types.payment` module provides the
[`PaymentCardNumber`][pydantic_extra_types.payment.PaymentCardNumber] data type.
"""

from __future__ import annotations

from enum import Enum
from typing import Any, ClassVar

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


class PaymentCardBrand(str, Enum):
    """Payment card brands supported by the [`PaymentCardNumber`][pydantic_extra_types.payment.PaymentCardNumber]."""

    amex = 'American Express'
    mastercard = 'Mastercard'
    visa = 'Visa'
    mir = 'Mir'
    maestro = 'Maestro'
    discover = 'Discover'
    verve = 'Verve'
    dankort = 'Dankort'
    troy = 'Troy'
    unionpay = 'UnionPay'
    jcb = 'JCB'
    other = 'other'

    def __str__(self) -> str:
        return self.value


class PaymentCardNumber(str):
    """A [payment card number](https://en.wikipedia.org/wiki/Payment_card_number)."""

    strip_whitespace: ClassVar[bool] = True
    """Whether to strip whitespace from the input value."""
    min_length: ClassVar[int] = 12
    """The minimum length of the card number."""
    max_length: ClassVar[int] = 19
    """The maximum length of the card number."""
    bin: str
    """The first 6 digits of the card number."""
    last4: str
    """The last 4 digits of the card number."""
    brand: PaymentCardBrand
    """The brand of the card."""

    def __init__(self, card_number: str):
        self.validate_digits(card_number)

        card_number = self.validate_luhn_check_digit(card_number)

        self.bin = card_number[:6]
        self.last4 = card_number[-4:]
        self.brand = self.validate_brand(card_number)

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls.validate,
            core_schema.str_schema(
                min_length=cls.min_length, max_length=cls.max_length, strip_whitespace=cls.strip_whitespace
            ),
        )

    @classmethod
    def validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> PaymentCardNumber:
        """Validate the `PaymentCardNumber` instance.

        Args:
            __input_value: The input value to validate.
            _: The validation info.

        Returns:
            The validated `PaymentCardNumber` instance.
        """
        return cls(__input_value)

    @property
    def masked(self) -> str:
        """The masked card number."""
        num_masked = len(self) - 10  # len(bin) + len(last4) == 10
        return f'{self.bin}{"*" * num_masked}{self.last4}'

    @classmethod
    def validate_digits(cls, card_number: str) -> None:
        """Validate that the card number is all digits.

        Args:
            card_number: The card number to validate.

        Raises:
            PydanticCustomError: If the card number is not all digits.
        """
        if not card_number or not all('0' <= c <= '9' for c in card_number):
            raise PydanticCustomError('payment_card_number_digits', 'Card number is not all digits')

    @classmethod
    def validate_luhn_check_digit(cls, card_number: str) -> str:
        """Validate the payment card number.
        Based on the [Luhn algorithm](https://en.wikipedia.org/wiki/Luhn_algorithm).

        Args:
            card_number: The card number to validate.

        Returns:
            The validated card number.

        Raises:
            PydanticCustomError: If the card number is not valid.
        """
        sum_ = int(card_number[-1])
        length = len(card_number)
        parity = length % 2
        for i in range(length - 1):
            digit = int(card_number[i])
            if i % 2 == parity:
                digit *= 2
            if digit > 9:
                digit -= 9
            sum_ += digit
        valid = sum_ % 10 == 0
        if not valid:
            raise PydanticCustomError('payment_card_number_luhn', 'Card number is not luhn valid')
        return card_number

    @staticmethod
    def validate_brand(card_number: str) -> PaymentCardBrand:
        """Validate length based on
        [BIN](https://en.wikipedia.org/wiki/Payment_card_number#Issuer_identification_number_(IIN))
        for major brands.

        Args:
            card_number: The card number to validate.

        Returns:
            The validated card brand.

        Raises:
            PydanticCustomError: If the card number is not valid.
        """
        brand = PaymentCardBrand.other

        if card_number[0] == '4':
            brand = PaymentCardBrand.visa
            required_length = [13, 16, 19]
        elif 51 <= int(card_number[:2]) <= 55:
            brand = PaymentCardBrand.mastercard
            required_length = [16]
        elif card_number[:2] in {'34', '37'}:
            brand = PaymentCardBrand.amex
            required_length = [15]
        elif 2200 <= int(card_number[:4]) <= 2204:
            brand = PaymentCardBrand.mir
            required_length = list(range(16, 20))
        elif card_number[:4] in {'5018', '5020', '5038', '5893', '6304', '6759', '6761', '6762', '6763'} or card_number[
            :6
        ] in (
            '676770',
            '676774',
        ):
            brand = PaymentCardBrand.maestro
            required_length = list(range(12, 20))
        elif card_number.startswith('65') or 644 <= int(card_number[:3]) <= 649 or card_number.startswith('6011'):
            brand = PaymentCardBrand.discover
            required_length = list(range(16, 20))
        elif (
            506099 <= int(card_number[:6]) <= 506198
            or 650002 <= int(card_number[:6]) <= 650027
            or 507865 <= int(card_number[:6]) <= 507964
        ):
            brand = PaymentCardBrand.verve
            required_length = [16, 18, 19]
        elif card_number[:4] in {'5019', '4571'}:
            brand = PaymentCardBrand.dankort
            required_length = [16]
        elif card_number.startswith('9792'):
            brand = PaymentCardBrand.troy
            required_length = [16]
        elif card_number[:2] in {'62', '81'}:
            brand = PaymentCardBrand.unionpay
            required_length = [16, 19]
        elif 3528 <= int(card_number[:4]) <= 3589:
            brand = PaymentCardBrand.jcb
            required_length = [16, 19]

        valid = len(card_number) in required_length if brand != PaymentCardBrand.other else True

        if not valid:
            raise PydanticCustomError(
                'payment_card_number_brand',
                f'Length for a {brand} card must be {" or ".join(map(str, required_length))}',
                {'brand': brand, 'required_length': required_length},
            )

        return brand



from string import printable

import pytest
from pydantic import BaseModel, ValidationError

from pydantic_extra_types.country import (
    CountryAlpha2,
    CountryAlpha3,
    CountryInfo,
    CountryNumericCode,
    CountryShortName,
    _index_by_alpha2,
    _index_by_alpha3,
    _index_by_numeric_code,
    _index_by_short_name,
)

PARAMS_AMOUNT = 20


@pytest.fixture(scope='module', name='ProductAlpha2')
def product_alpha2_fixture():
    class Product(BaseModel):
        made_in: CountryAlpha2

    return Product


@pytest.fixture(scope='module', name='ProductAlpha3')
def product_alpha3_fixture():
    class Product(BaseModel):
        made_in: CountryAlpha3

    return Product


@pytest.fixture(scope='module', name='ProductShortName')
def product_short_name_fixture():
    class Product(BaseModel):
        made_in: CountryShortName

    return Product


@pytest.fixture(scope='module', name='ProductNumericCode')
def product_numeric_code_fixture():
    class Product(BaseModel):
        made_in: CountryNumericCode

    return Product


@pytest.mark.parametrize('alpha2, country_data', list(_index_by_alpha2().items())[:PARAMS_AMOUNT])
def test_valid_alpha2(alpha2: str, country_data: CountryInfo, ProductAlpha2):
    banana = ProductAlpha2(made_in=alpha2)
    assert banana.made_in == country_data.alpha2
    assert banana.made_in.alpha3 == country_data.alpha3
    assert banana.made_in.numeric_code == country_data.numeric_code
    assert banana.made_in.short_name == country_data.short_name


@pytest.mark.parametrize('alpha2', list(printable))
def test_invalid_alpha2(alpha2: str, ProductAlpha2):
    with pytest.raises(ValidationError, match='Invalid country alpha2 code'):
        ProductAlpha2(made_in=alpha2)


@pytest.mark.parametrize('alpha3, country_data', list(_index_by_alpha3().items())[:PARAMS_AMOUNT])
def test_valid_alpha3(alpha3: str, country_data: CountryInfo, ProductAlpha3):
    banana = ProductAlpha3(made_in=alpha3)
    assert banana.made_in == country_data.alpha3
    assert banana.made_in.alpha2 == country_data.alpha2
    assert banana.made_in.numeric_code == country_data.numeric_code
    assert banana.made_in.short_name == country_data.short_name


@pytest.mark.parametrize('alpha3', list(printable))
def test_invalid_alpha3(alpha3: str, ProductAlpha3):
    with pytest.raises(ValidationError, match='Invalid country alpha3 code'):
        ProductAlpha3(made_in=alpha3)


@pytest.mark.parametrize('short_name, country_data', list(_index_by_short_name().items())[:PARAMS_AMOUNT])
def test_valid_short_name(short_name: str, country_data: CountryInfo, ProductShortName):
    banana = ProductShortName(made_in=short_name)
    assert banana.made_in == country_data.short_name
    assert banana.made_in.alpha2 == country_data.alpha2
    assert banana.made_in.alpha3 == country_data.alpha3
    assert banana.made_in.numeric_code == country_data.numeric_code


@pytest.mark.parametrize('short_name', list(printable))
def test_invalid_short_name(short_name: str, ProductShortName):
    with pytest.raises(ValidationError, match='Invalid country short name'):
        ProductShortName(made_in=short_name)


@pytest.mark.parametrize('numeric_code, country_data', list(_index_by_numeric_code().items())[:PARAMS_AMOUNT])
def test_valid_numeric_code(numeric_code: str, country_data: CountryInfo, ProductNumericCode):
    banana = ProductNumericCode(made_in=numeric_code)
    assert banana.made_in == country_data.numeric_code
    assert banana.made_in.alpha2 == country_data.alpha2
    assert banana.made_in.alpha3 == country_data.alpha3
    assert banana.made_in.short_name == country_data.short_name


@pytest.mark.parametrize('numeric_code', list(printable))
def test_invalid_numeric_code(numeric_code: str, ProductNumericCode):
    with pytest.raises(ValidationError, match='Invalid country numeric code'):
        ProductNumericCode(made_in=numeric_code)



"""
The `pydantic_extra_types.routing_number` module provides the
[`ABARoutingNumber`][pydantic_extra_types.routing_number.ABARoutingNumber] data type.
"""
from typing import Any, ClassVar, Type

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


class ABARoutingNumber(str):
    """The `ABARoutingNumber` data type is a string of 9 digits representing an ABA routing transit number.

    The algorithm used to validate the routing number is described in the
    [ABA routing transit number](https://en.wikipedia.org/wiki/ABA_routing_transit_number#Check_digit)
    Wikipedia article.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.routing_number import ABARoutingNumber

    class BankAccount(BaseModel):
        routing_number: ABARoutingNumber

    account = BankAccount(routing_number='122105155')
    print(account)
    #> routing_number='122105155'
    ```
    """

    strip_whitespace: ClassVar[bool] = True
    min_length: ClassVar[int] = 9
    max_length: ClassVar[int] = 9

    def __init__(self, routing_number: str):
        self._validate_digits(routing_number)
        self._routing_number = self._validate_routing_number(routing_number)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(
                min_length=cls.min_length,
                max_length=cls.max_length,
                strip_whitespace=cls.strip_whitespace,
                strict=False,
            ),
        )

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> 'ABARoutingNumber':
        return cls(__input_value)

    @classmethod
    def _validate_digits(cls, routing_number: str) -> None:
        """Check that the routing number is all digits.

        Args:
            routing_number: The routing number to validate.

        Raises:
            PydanticCustomError: If the routing number is not all digits.
        """
        if not routing_number.isdigit():
            raise PydanticCustomError('aba_routing_number', 'routing number is not all digits')

    @classmethod
    def _validate_routing_number(cls, routing_number: str) -> str:
        """Check [digit algorithm](https://en.wikipedia.org/wiki/ABA_routing_transit_number#Check_digit) for
        [ABA routing transit number](https://www.routingnumber.com/).

        Args:
            routing_number: The routing number to validate.

        Raises:
            PydanticCustomError: If the routing number is incorrect.
        """
        checksum = (
            3 * (sum(map(int, [routing_number[0], routing_number[3], routing_number[6]])))
            + 7 * (sum(map(int, [routing_number[1], routing_number[4], routing_number[7]])))
            + sum(map(int, [routing_number[2], routing_number[5], routing_number[8]]))
        )
        if checksum % 10 != 0:
            raise PydanticCustomError('aba_routing_number', 'Incorrect ABA routing transit number')
        return routing_number



"""
The `pydantic_extra_types.isbn` module provides functionality to recieve and validate ISBN
(International Standard Book Number) in 10-digit and 13-digit formats. The output is always ISBN-13.
"""

from __future__ import annotations

from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


def isbn10_digit_calc(isbn: str) -> str:
    """
    Calc a ISBN-10 last digit from the provided str value. More information of validation algorithm on [Wikipedia](https://en.wikipedia.org/wiki/ISBN#Check_digits)

    Args:
        isbn: The str value representing the ISBN in 10 digits.

    Returns:
        The calculated last digit.
    """
    total = sum(int(digit) * (10 - idx) for idx, digit in enumerate(isbn[:9]))

    for check_digit in range(1, 11):
        if (total + check_digit) % 11 == 0:
            valid_check_digit = 'X' if check_digit == 10 else str(check_digit)

    return valid_check_digit


def isbn13_digit_calc(isbn: str) -> str:
    """
    Calc a ISBN-13 last digit from the provided str value. More information of validation algorithm on [Wikipedia](https://en.wikipedia.org/wiki/ISBN#Check_digits)

    Args:
        isbn: The str value representing the ISBN in 13 digits.

    Returns:
        The calculated last digit.
    """
    total = sum(int(digit) * (1 if idx % 2 == 0 else 3) for idx, digit in enumerate(isbn[:12]))

    check_digit = (10 - (total % 10)) % 10

    return str(check_digit)


class ISBN(str):
    """Represents a ISBN and provides methods for conversion, validation, and serialization.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.isbn import ISBN


    class Book(BaseModel):
        isbn: ISBN

    book = Book(isbn="8537809667")
    print(book)
    #> isbn='9788537809662'
    ```
    """

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.with_info_before_validator_function(
            cls._validate,
            core_schema.str_schema(),
        )

    @classmethod
    def _validate(cls, __input_value: str, _: Any) -> str:
        cls.validate_isbn_format(__input_value)

        return cls.convert_isbn10_to_isbn13(__input_value)

    @staticmethod
    def validate_isbn_format(value: str) -> None:
        """
        Validate a ISBN format from the provided str value.

        Args:
            value: The str value representing the ISBN in 10 or 13 digits.

        Raises:
            PydanticCustomError: If the value is not a valid ISBN.
        """

        isbn_length = len(value)

        if isbn_length not in (10, 13):
            raise PydanticCustomError('isbn_length', f'Length for ISBN must be 10 or 13 digits, not {isbn_length}')

        if isbn_length == 10:
            if not value[:-1].isdigit() or ((value[-1] != 'X') and (not value[-1].isdigit())):
                raise PydanticCustomError('isbn10_invalid_characters', 'First 9 digits of ISBN-10 must be integers')
            if isbn10_digit_calc(value) != value[-1]:
                raise PydanticCustomError('isbn_invalid_digit_check_isbn10', 'Provided digit is invalid for given ISBN')

        if isbn_length == 13:
            if not value.isdigit():
                raise PydanticCustomError('isbn13_invalid_characters', 'All digits of ISBN-13 must be integers')
            if value[:3] not in ('978', '979'):
                raise PydanticCustomError(
                    'isbn_invalid_early_characters', 'The first 3 digits of ISBN-13 must be 978 or 979'
                )
            if isbn13_digit_calc(value) != value[-1]:
                raise PydanticCustomError('isbn_invalid_digit_check_isbn13', 'Provided digit is invalid for given ISBN')

    @staticmethod
    def convert_isbn10_to_isbn13(value: str) -> str:
        """
        Convert an ISBN-10 to ISBN-13.

        Args:
            value: The str value representing the ISBN.

        Returns:
            The converted ISBN or the original value if no conversion is necessary.
        """

        if len(value) == 10:
            base_isbn = f'978{value[:-1]}'
            isbn13_digit = isbn13_digit_calc(base_isbn)
            return ISBN(f'{base_isbn}{isbn13_digit}')

        return ISBN(value)
